/* D2Q9 lattice boltzmann functions */

#include <e-lib.h>
#include "../shared.h"

/* core indices */
extern unsigned int row, col, core;

/* ================================================================== */

/* velocities */
static const int d2q9_v[9][2] = { { 0, 0},
	{-1, 1}, {-1, 0}, {-1,-1}, { 0,-1},
	{ 1,-1}, { 1, 0}, { 1, 1}, { 0, 1},
};

/* weights */
static const FLOAT d2q9_w[9] = { 4./9.,
	1./36., 1./9., 1./36., 1./9.,
	1./36., 1./9., 1./36., 1./9.,
};

/* ================================================================== */

void init(block_t f)
{
	/* all with rho = 0.1 */
	for(int y = 0; y < NODES_Y; y++)
		for(int x = 0; x < NODES_X; x++)
			for(int q = 0; q < 9; q++)
				f[y][x][q] = 0.1 * d2q9_w[q];

	if(core == 0) {
		/* except here with 0.2 */
		for(int q = 0; q < 9; q++)
			f[0][0][q] = 0.2 * d2q9_w[q];
	}

	return;
}

void collide(block_t f, int x, int y, FLOAT omega)
{
	if(row == 0 && y == 0) {
		/* Zou/He boundary at top, with velocity */
		const FLOAT UX = 0.0;		/* speed in x-direction */
		const FLOAT UY = 0.0;		/* speed in y-direction */

		FLOAT rho = ( f[y][x][0] + f[y][x][6] + f[y][x][2] +
		        2 * ( f[y][x][5] + f[y][x][3] + f[y][x][4] ) ) /
		        (1 - UY);

		FLOAT tmp1 = ( f[y][x][6] - f[y][x][2] ) / 2;
		FLOAT tmp2 = rho * (-UX - UY / 3) / 2;

		f[y][x][7] = f[y][x][3] - tmp1 - tmp2;
		f[y][x][1] = f[y][x][5] + tmp1 + tmp2;
		f[y][x][8] = f[y][x][4] + 2 * rho * UY / 3;

	} else if(row == CORES_Y-1 && y == NODES_Y-1) {
		/* Zou/He boundary at bottom, no velocity */
		FLOAT tmp = ( f[y][x][6] - f[y][x][2] ) / 2;
		f[y][x][3] = f[y][x][7] + tmp;
		f[y][x][5] = f[y][x][1] - tmp;
		f[y][x][4] = f[y][x][8];

	} else if(col == 0 && x == 0) {
		/* Zou/He boundary at left, no velocity */
		FLOAT tmp = ( f[y][x][8] - f[y][x][4] ) / 2;
		f[y][x][5] = f[y][x][1] + tmp;
		f[y][x][7] = f[y][x][3] - tmp;
		f[y][x][6] = f[y][x][2];

	} else if(col == CORES_X-1 && x == NODES_X-1) {
		/* Zou/He boundary at right, no velocity */
		FLOAT tmp = ( f[y][x][8] - f[y][x][4] ) / 2;
		f[y][x][1] = f[y][x][5] - tmp;
		f[y][x][3] = f[y][x][7] + tmp;
		f[y][x][2] = f[y][x][6];
	}

	/* macroscopic */
	FLOAT rho = f[y][x][0] + f[y][x][1] + f[y][x][2] +
		f[y][x][3] + f[y][x][4] + f[y][x][5] +
		f[y][x][6] + f[y][x][7] + f[y][x][8];
	FLOAT ux = (f[y][x][7] + f[y][x][6] + f[y][x][5] -
		f[y][x][1] - f[y][x][2] - f[y][x][3]) / rho;
	FLOAT uy = (f[y][x][1] + f[y][x][8] + f[y][x][7] -
		f[y][x][3] - f[y][x][4] - f[y][x][5]) / rho;
	FLOAT sqr = 1.5 * (ux*ux + uy*uy);

	/* update node */
	for(int q = 0; q < 9; q++) {
		FLOAT cu = ux*d2q9_v[q][0] + uy*d2q9_v[q][1];
		FLOAT eq = rho * d2q9_w[q] *
			(1. + 3. * cu + 4.5 * cu*cu - sqr);
		f[y][x][q] *= (1.0 - omega);
		f[y][x][q] += omega * eq;
	}

	/* swap */
	for(int q = 1; q <= 4; q++) {
		FLOAT tmp    = f[y][x][q];
		f[y][x][q]   = f[y][x][q+4];
		f[y][x][q+4] = tmp;
	}
}

void stream(block_t f, int x, int y)
{
	for(int q = 1; q <= 4; q++) {
		int next_row = row;
		int next_col = col;
		int next_x   = x + d2q9_v[q][0];
		int next_y   = y + d2q9_v[q][1];

		/* inner borders (extend) */
		if(next_x < 0)             { next_col--; next_x += NODES_X; }
		else if(next_x >= NODES_X) { next_col++; next_x -= NODES_X; }
		if(next_y < 0)             { next_row--; next_y += NODES_Y; }
		else if(next_y >= NODES_Y) { next_row++; next_y -= NODES_Y; }

#if 0
		/* outer borders (wrap around) */
		if(next_col < 0)             { next_col += CORES_X; }
		else if(next_col >= CORES_X) { next_col -= CORES_X; }
		if(next_row < 0)             { next_row += CORES_Y; }
		else if(next_row >= CORES_Y) { next_row -= CORES_Y; }
#else
		/* don't stream from outside the domain */
		if(next_col < 0)             { continue; }
		else if(next_col >= CORES_X) { continue; }
		if(next_row < 0)             { continue; }
		else if(next_row >= CORES_Y) { continue; }
#endif

		/* f: local block, g: local or remote block */
		block_t *g = (void*)f;
		if(next_row != row || next_col != col) {
			g = e_get_global_address(next_col, next_row, (void*)f);
		}

		/* stream/swap f and g */
		FLOAT tmp    = f[y][x][q+4];
		f[y][x][q+4] = (*g)[next_y][next_x][q];
		(*g)[next_y][next_x][q] = tmp;
	}
}

void bulk(block_t f, FLOAT omega)
{
	/* don't touch the border nodes */
	for(int x = 1; x < NODES_X-1; x++) {
		for(int y = 1; y < NODES_Y-1; y++) {
			/* macroscopic */
			FLOAT rho = f[y][x][0] + f[y][x][1] + f[y][x][2] +
				f[y][x][3] + f[y][x][4] + f[y][x][5] +
				f[y][x][6] + f[y][x][7] + f[y][x][8];
			FLOAT ux = (f[y][x][7] + f[y][x][6] + f[y][x][5] -
				f[y][x][1] - f[y][x][2] - f[y][x][3]) / rho;
			FLOAT uy = (f[y][x][1] + f[y][x][8] + f[y][x][7] -
				f[y][x][3] - f[y][x][4] - f[y][x][5]) / rho;
			FLOAT sqr = 1.5 * (ux*ux + uy*uy);

			/* update node */
			for(int q = 0; q < 9; q++) {
				FLOAT cu = ux*d2q9_v[q][0] + uy*d2q9_v[q][1];
				FLOAT eq = rho * d2q9_w[q] *
					(1. + 3. * cu + 4.5 * cu*cu - sqr);
				f[y][x][q] *= (1.0 - omega);
				f[y][x][q] += omega * eq;
			}

			/* stream */
			for(int q = 0; q <= 4; q++) {
				int next_x = x + d2q9_v[q][0];
				int next_y = y + d2q9_v[q][1];

				FLOAT tmp    = f[y][x][q];
				f[y][x][q]   = f[y][x][q+4];
				f[y][x][q+4] = f[next_y][next_x][q];
				f[next_y][next_x][q] = tmp;
			}
		}
	}
}

