/* D2Q9 lattice boltzmann implementation */

#include <e-lib.h>
#include "../shared.h"

#include <stdint.h>
#include <string.h>

#include "d2q9.h"

/* shared memory overlay */
volatile shm_t shm SECTION(".shared_dram");

/* statically allocate dummy memory and local block overlay
   to prevent linker from putting stuff in banks 1..3 */
static uint8_t dummy_bank1[8192] USED SECTION(".data_bank1");
static uint8_t dummy_bank2[8192] USED SECTION(".data_bank2");
static uint8_t dummy_bank3[8192] USED SECTION(".data_bank3");
static block_t *block = (void*)0x2000;

/* timer values */
times_t times = {0};

/* barrier structures */
volatile e_barrier_t  barriers[CORES];
         e_barrier_t *tgt_bars[CORES];

/* global index variables */
unsigned int row, col, core;

#define READ_TIMER(X) \
	do { \
		times[X] = E_CTIMER_MAX - e_ctimer_stop(E_CTIMER_0); \
		e_ctimer_set(E_CTIMER_0, E_CTIMER_MAX); \
		e_ctimer_start(E_CTIMER_0, E_CTIMER_CLK); \
	} while(0);

int main()
{
	const FLOAT omega = 1.0;

	/* compile-time checks */
	BUILD_BUG(NODES * sizeof(node_t) > 24 * 1024);
	BUILD_BUG(NODES_X < 3 || NODES_Y < 3);
	BUILD_BUG(CORES_X < 1 || CORES_Y < 1);
	BUILD_BUG(CORES_X > 8 || CORES_Y > 8);

	/* save mesh coordinates */
	e_coords_from_coreid(e_get_coreid(), &col, &row);
	core = row * CORES_X + col;

	/* initialize barrier */
	e_barrier_init(barriers, tgt_bars);

	/* initialize block */
	init(*block);

	/* main loop */
	for(int iter = 0; iter < 500; iter++) {
		READ_TIMER(0);

#if 1
		/* collide all nodes */
		for(int y = 0; y < NODES_Y; y++)
			for(int x = 0; x < NODES_X; x++)
				collide(*block, x, y, omega);

		/* synchronize */
		READ_TIMER(1);
		e_barrier(barriers, tgt_bars);
		READ_TIMER(2);

		/* stream all nodes */
		for(int y = 0; y < NODES_Y; y++)
			for(int x = 0; x < NODES_X; x++)
				stream(*block, x, y);
		READ_TIMER(3);

#else
		/* collide boundaries: top, bottom */
		for(int x = 0; x < NODES_X; x++) {
			collide(*block, x, 0,          omega);
			collide(*block, x, NODES_Y-1, omega);
		}
		READ_TIMER(1);

		/* collide boundaries: left, right */
		for(int y = 1; y < NODES_Y-1; y++) {
			collide(*block, 0,         y, omega);
			collide(*block, NODES_X-1, y, omega);
		}

		/* synchronize */
		READ_TIMER(2);
		e_barrier(barriers, tgt_bars);
		READ_TIMER(3);

		/* collide and stream the bulk */
		bulk(*block, omega);
		READ_TIMER(4);

		/* stream the boundaries: top, bottom */
		for(int x = 0; x < NODES_X; x++) {
			stream(*block, x, 0         );
			stream(*block, x, NODES_Y-1);
		}
		READ_TIMER(5);

		/* stream the boundaries: left, right */
		for(int y = 1; y < NODES_Y-1; y++) {
			stream(*block, 0,          y);
			stream(*block, NODES_X-1, y);
		}
		READ_TIMER(6);
#endif

		/* copy data to shm if necessary */
		if(!(iter % 1)) {
			/* copy lattice to shm */
			memcpy(&shm.lattice[row][col], block, sizeof(block_t));

			/* copy times to shm */
			memcpy(&shm.times[row][col], times, sizeof(times_t));

			/* synchronize */
			e_barrier(barriers, tgt_bars);

			/* core 0: write counter and flag host; wait */
			if(core == 0) {
				shm.iteration = iter;
				shm.pollflag  = POLL_READY;
				while(shm.pollflag == POLL_READY);
			}
		}

		/* synchronize */
		e_barrier(barriers, tgt_bars);
	}

	/* last iteration done: flag host and stop */
	if(core == 0) shm.pollflag = POLL_DONE;
	while(1) __asm__ volatile("idle");
}

