/* Helper Functions to handle data (2D) */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>

#include "../shared.h"

/* fix file owner if run with sudo */
void fixsudo(const char *filename)
{
	if(getenv("SUDO_UID") && getenv("SUDO_GID")) {
		int uid = atoi(getenv("SUDO_UID"));
		int gid = atoi(getenv("SUDO_GID"));
		if(chown(filename, uid, gid)) {
			perror("fixsudo/chown");
			return;
		}
	}
}

/* write a (semi-) human-readable dump of the lattice */
void write_populations(block_t f[CORES_Y][CORES_X], int iter)
{
	FILE *file = fopen("populations.dat", "a");
	if(!file) {
		perror("write_populations/fopen");
		return;
	}

	for(int cy = 0; cy < CORES_Y; cy++) {
		for(int y = 0; y < NODES_Y; y++) {
			for(int cx = 0; cx < CORES_X; cx++) {
				for(int x = 0; x < NODES_X; x++) {
					fprintf(file, "%3d: [%3d,%3d]: ",
						iter,
						cx * NODES_X + x,
						cy * NODES_Y + y
					);
					for(int q = 0; q < 9; q++) {
						fprintf(file, "%.5f ",
							f[cy][cx][y][x][q]);
					}
					fprintf(file, "\n");
				}
			}
		}
	}
	fprintf(file, "\n");

	/* close */
	fclose(file);

	return;
}

/* write an 8-bit grayscale, binary PPM image of the particle density */
void write_density(block_t f[CORES_Y][CORES_X], int iter)
{
	char name[32]; snprintf(name, 32, "./tmp/i%06d.ppm", iter);

	/* open image file and write header */
	FILE *file = fopen(name, "wb");
	if(!file) {
		perror("write_density/fopen");
		return;
	}
	fprintf(file, "P5\n%d %d\n%d\n",
		CORES_X*NODES_X, CORES_Y*NODES_Y, 255);

	/* calculate all densities and remember min/max */
	FLOAT min = 1.0, max = 0;
	FLOAT rhos[CORES_Y][NODES_Y][CORES_X][NODES_X];
	for(int cy = 0; cy < CORES_Y; cy++) {
		for(int y = 0; y < NODES_Y; y++) {
			for(int cx = 0; cx < CORES_X; cx++) {
				for(int x = 0; x < NODES_X; x++) {
					FLOAT rho = 0;
					for(int q = 0; q < 9; q++)
						rho += f[cy][cx][y][x][q];
					rhos[cy][y][cx][x] = rho;

					if(rho < min) min = rho;
					if(rho > max) max = rho;
				}
			}
		}
	}

	/* scale values and write them to the image */
	for(int cy = 0; cy < CORES_Y; cy++) {
		for(int y = 0; y < NODES_Y; y++) {
			for(int cx = 0; cx < CORES_X; cx++) {
				for(int x = 0; x < NODES_X; x++) {
					unsigned char gray;
					gray = (255. * (rhos[cy][y][cx][x]-min) / (max-min));
					fwrite(&gray, 1, 1, file);
				}
			}
		}
	}

	/* close the file and chown if run with sudo */
	fclose(file);
	fixsudo(name);

	return;
}

/* write an 8-bit grayscale, binary PPM image of the particle velocity */
void write_velocity(block_t f[CORES_Y][CORES_X], int iter)
{
	char name[32]; snprintf(name, 32, "./tmp/i%06d.ppm", iter);

	/* open image file and write header */
	FILE *file = fopen(name, "wb");
	if(!file) {
		perror("write_velocity/fopen");
		return;
	}
	fprintf(file, "P5\n%d %d\n%d\n",
		CORES_X*NODES_X, CORES_Y*NODES_Y, 255);


	/* calculate all velocities and remember min/max */
	FLOAT min = 1000, max = 0;
	FLOAT us[CORES_Y][NODES_Y][CORES_X][NODES_X];
	for(int cy = 0; cy < CORES_Y; cy++) {
		for(int y = 0; y < NODES_Y; y++) {
			for(int cx = 0; cx < CORES_X; cx++) {
				for(int x = 0; x < NODES_X; x++) {
					FLOAT rho = (
						f[cy][cx][y][x][0] +
						f[cy][cx][y][x][1] +
						f[cy][cx][y][x][2] +
						f[cy][cx][y][x][3] +
						f[cy][cx][y][x][4] +
						f[cy][cx][y][x][5] +
						f[cy][cx][y][x][6] +
						f[cy][cx][y][x][7] +
						f[cy][cx][y][x][8]
					);

					FLOAT ux = (
						f[cy][cx][y][x][5] +
						f[cy][cx][y][x][6] +
						f[cy][cx][y][x][7] -
						f[cy][cx][y][x][1] -
						f[cy][cx][y][x][2] -
						f[cy][cx][y][x][3]
					) / rho;

					FLOAT uy = (
						f[cy][cx][y][x][1] +
						f[cy][cx][y][x][7] +
						f[cy][cx][y][x][8] -
						f[cy][cx][y][x][3] -
						f[cy][cx][y][x][4] -
						f[cy][cx][y][x][5]
					) / rho;

					FLOAT u = sqrtf(ux*ux + uy*uy);
					if(u < min) min = u;
					if(u > max) max = u;
					us[cy][y][cx][x] = u;
				}
			}
		}
	}

	/* scale values and write them to the image */
	for(int cy = 0; cy < CORES_Y; cy++) {
		for(int y = 0; y < NODES_Y; y++) {
			for(int cx = 0; cx < CORES_X; cx++) {
				for(int x = 0; x < NODES_X; x++) {
					unsigned char gray;
					gray = (255. * (us[cy][y][cx][x]-min) / (max-min));
					fwrite(&gray, 1, 1, file);
				}
			}
		}
	}

	/* close the file and chown if run with sudo */
	fclose(file);
	fixsudo(name);

	return;
}

/* convert image files to animated gif ./tmp/anim.gif */
void convert_to_gif(void)
{
	int result;

	/* call imagemagick */
	result = system("convert ./tmp/i*.ppm ./tmp/anim.gif"); (void)result;
	fixsudo("./tmp/anim.gif");

	return;
}

/* convert image files to mp4 ./tmp/anim.mp4 */
void convert_to_mp4(void)
{
	int result;

	/* call ffmpeg
	   FIXME: requires ffmpeg binary with '-pattern_type glob' support,
	          so this will not work on the Parallella without updates! */
	result = system("ffmpeg -v quiet -pattern_type glob -i "
	                "'./tmp/i*.ppm' ./tmp/anim.mp4"); (void)result;
	fixsudo("./tmp/anim.mp4");

	return;
}

/* write timer values */
void write_timers(times_t times[CORES_Y][CORES_X], uint32_t iter)
{
	FILE *file = fopen("timers.dat", "ab");
	if(!file) {
		perror("write_timers/fopen");
		return;
	}

	fprintf(file, "Timers: i=%d\n", iter);
	for(int y = 0; y < CORES_Y; y++) {
		for(int x = 0; x < CORES_X; x++) {
			fprintf(file, "[%d,%d]: ", x, y);
			for(int t = 0; t < TIMERS; t++) {
				fprintf(file, "%8d ", times[y][x][t]);
			}
			fprintf(file, "\n");
		}
	}

	fclose(file);
}
