/* Lattice Boltzmann Algorithm Implementation (D3Q19) */

#include <e-lib.h>
#include "../shared.h"

/* core indices */
extern unsigned int row, col, core;

/* ================================================================== */

static const int d3q19_v[19][3] = { { 0, 0, 0},
	{-1, 0, 0}, { 0,-1, 0}, { 0, 0,-1}, {-1,-1, 0}, {-1, 1, 0}, {-1, 0,-1},
	{-1, 0, 1}, { 0,-1,-1}, { 0,-1, 1}, { 1, 0, 0}, { 0, 1, 0}, { 0, 0, 1},
	{ 1, 1, 0}, { 1,-1, 0}, { 1, 0, 1}, { 1, 0,-1}, { 0, 1, 1}, { 0, 1,-1},
};
static const FLOAT d3q19_w[19] = { 1./3.,
	1./18., 1./18., 1./18., 1./36., 1./36., 1./36.,
	1./36., 1./36., 1./36., 1./18., 1./18., 1./18.,
	1./36., 1./36., 1./36., 1./36., 1./36., 1./36.,
};

/* ================================================================== */

void init(block_t f)
{
	/* initialize all nodes with rho = 0.1 */
	for(int z = 0; z < NODES_Z; z++)
		for(int y = 0; y < NODES_Y; y++)
			for(int x = 0; x < NODES_X; x++)
				for(int q = 0; q < 19; q++)
					f[z][y][x][q] = 0.1 * d3q19_w[q];

#if 0
	/* except here */
	if(core == 0)
		for(int q = 0; q < 19; q++)
			f[1][1][1][q] = 0.2 * d3q19_w[q];
#endif
}

/* ================================================================== */

void collide(block_t f, int x, int y, int z, FLOAT omega)
{
#if 1
	/* Zou/He boundary conditions */
	if(row == 0 && z == 0) {				/* top */
		const FLOAT Ux = 0.6;	/* wall speed */
		const FLOAT Uy = 0.0;
		const FLOAT Uz = 0.0;

		FLOAT rho = ( f[z][y][x][0]  + f[z][y][x][1]  + f[z][y][x][2]  +
		              f[z][y][x][4]  + f[z][y][x][5]  + f[z][y][x][10] +
		              f[z][y][x][11] + f[z][y][x][13] + f[z][y][x][14] +
		        2 * ( f[z][y][x][16] + f[z][y][x][18] + f[z][y][x][3]  +
		              f[z][y][x][6]  + f[z][y][x][8]  ) / (1 + Uz) );

		FLOAT tmp0 = rho / 2;
		FLOAT tmp1 = ( f[z][y][x][10] - f[z][y][x][1] ) / 2;
		FLOAT tmp2 = ( f[z][y][x][11] - f[z][y][x][2] ) / 2;
		FLOAT tmp3 = ( f[z][y][x][13] - f[z][y][x][4] ) / 2;
		FLOAT tmp4 = ( f[z][y][x][14] - f[z][y][x][5] ) / 2;

		f[z][y][x][15] = f[z][y][x][6]  - tmp1 - tmp3 - tmp4 -
			tmp0 * (Uz/3 - Ux);
		f[z][y][x][7]  = f[z][y][x][16] + tmp1 + tmp3 + tmp4 -
			tmp0 * (Uz/3 + Ux);
		f[z][y][x][17] = f[z][y][x][8]  - tmp2 - tmp3 + tmp4 -
			tmp0 * (Uz/3 - Uy);
		f[z][y][x][9]  = f[z][y][x][18] + tmp2 + tmp3 - tmp4 -
			tmp0 * (Uz/3 + Uy);
		f[z][y][x][12] = f[z][y][x][3] - rho * Uz / 3;

	} else if(row == CORES_Y-1 && z == NODES_Z-1) {		/* bottom */
		FLOAT tmp1 = ( f[z][y][x][10] - f[z][y][x][1] ) / 2;
		FLOAT tmp2 = ( f[z][y][x][11] - f[z][y][x][2] ) / 2;
		FLOAT tmp3 = ( f[z][y][x][13] - f[z][y][x][4] ) / 2;
		FLOAT tmp4 = ( f[z][y][x][14] - f[z][y][x][5] ) / 2;

		f[z][y][x][6]  = f[z][y][x][15] + tmp1 + tmp3 + tmp4;
		f[z][y][x][16] = f[z][y][x][7]  - tmp1 - tmp3 - tmp4;
		f[z][y][x][8]  = f[z][y][x][17] + tmp2 + tmp3 - tmp4;
		f[z][y][x][18] = f[z][y][x][9]  - tmp2 - tmp3 + tmp4;
		f[z][y][x][3]  = f[z][y][x][12];

	} else if(y == 0) {					/* front */
		FLOAT tmp1 = ( f[z][y][x][10] - f[z][y][x][1] ) / 2;
		FLOAT tmp2 = ( f[z][y][x][12] - f[z][y][x][3] ) / 2;
		FLOAT tmp3 = ( f[z][y][x][15] - f[z][y][x][6] ) / 2;
		FLOAT tmp4 = ( f[z][y][x][16] - f[z][y][x][7] ) / 2;

		f[z][y][x][13] = f[z][y][x][4]  - tmp1 - tmp3 - tmp4;
		f[z][y][x][5]  = f[z][y][x][14] + tmp1 + tmp3 + tmp4;
		f[z][y][x][17] = f[z][y][x][8]  - tmp2 - tmp3 - tmp4;
		f[z][y][x][18] = f[z][y][x][9]  + tmp2 + tmp3 + tmp4;
		f[z][y][x][11] = f[z][y][x][2];

	} else if(y == NODES_Y-1) {				/* back */
		FLOAT tmp1 = ( f[z][y][x][10] - f[z][y][x][1] ) / 2;
		FLOAT tmp2 = ( f[z][y][x][12] - f[z][y][x][3] ) / 2;
		FLOAT tmp3 = ( f[z][y][x][15] - f[z][y][x][6] ) / 2;
		FLOAT tmp4 = ( f[z][y][x][16] - f[z][y][x][7] ) / 2;

		f[z][y][x][4]  = f[z][y][x][13] + tmp1 + tmp3 + tmp4;
		f[z][y][x][14] = f[z][y][x][5]  - tmp1 - tmp3 - tmp4;
		f[z][y][x][8]  = f[z][y][x][17] + tmp2 + tmp3 + tmp4;
		f[z][y][x][9]  = f[z][y][x][18] - tmp2 - tmp3 - tmp4;
		f[z][y][x][2]  = f[z][y][x][11];

	} else if(col == 0 && x == 0) {				/* left */
		FLOAT tmp1 = ( f[z][y][x][11] - f[z][y][x][2] ) / 2;
		FLOAT tmp2 = ( f[z][y][x][12] - f[z][y][x][3] ) / 2;
		FLOAT tmp3 = ( f[z][y][x][17] - f[z][y][x][8] ) / 2;
		FLOAT tmp4 = ( f[z][y][x][18] - f[z][y][x][9] ) / 2;

		f[z][y][x][13] = f[z][y][x][4]  - tmp1 - tmp3 - tmp4;
		f[z][y][x][14] = f[z][y][x][5]  + tmp1 + tmp3 + tmp4;
		f[z][y][x][15] = f[z][y][x][6]  - tmp2 - tmp3 + tmp4;
		f[z][y][x][16] = f[z][y][x][7]  + tmp2 + tmp3 - tmp4;
		f[z][y][x][10] = f[z][y][x][1];

	} else if(col == CORES_X-1 && x == NODES_X-1) {		/* right */
		FLOAT tmp1 = ( f[z][y][x][11] - f[z][y][x][2] ) / 2;
		FLOAT tmp2 = ( f[z][y][x][12] - f[z][y][x][3] ) / 2;
		FLOAT tmp3 = ( f[z][y][x][17] - f[z][y][x][8] ) / 2;
		FLOAT tmp4 = ( f[z][y][x][18] - f[z][y][x][9] ) / 2;

		f[z][y][x][4]  = f[z][y][x][13] + tmp1 + tmp3 + tmp4;
		f[z][y][x][5]  = f[z][y][x][14] - tmp1 - tmp3 - tmp4;
		f[z][y][x][6]  = f[z][y][x][15] + tmp2 + tmp3 - tmp4;
		f[z][y][x][7]  = f[z][y][x][16] - tmp2 - tmp3 + tmp4;
		f[z][y][x][1]  = f[z][y][x][10];
	}
#endif

	/* macroscopic */
	FLOAT rho =
		f[z][y][x][0] + f[z][y][x][1] + f[z][y][x][2] +
		f[z][y][x][3] + f[z][y][x][4] + f[z][y][x][5] +
		f[z][y][x][6] + f[z][y][x][7] + f[z][y][x][8] +
		f[z][y][x][9] + f[z][y][x][10] + f[z][y][x][11] +
		f[z][y][x][12] + f[z][y][x][13] + f[z][y][x][14] +
		f[z][y][x][15] + f[z][y][x][16] + f[z][y][x][17] +
		f[z][y][x][18];

	FLOAT ux = (
		f[z][y][x][10] + f[z][y][x][13] + f[z][y][x][14] +
		f[z][y][x][15] + f[z][y][x][16] - f[z][y][x][1] -
		f[z][y][x][4]  - f[z][y][x][5]  - f[z][y][x][6] -
		f[z][y][x][7] ) / rho;

	FLOAT uy = (
		f[z][y][x][5]  + f[z][y][x][11] + f[z][y][x][13] +
		f[z][y][x][17] + f[z][y][x][18] - f[z][y][x][2]  -
		f[z][y][x][4]  - f[z][y][x][8]  - f[z][y][x][9]  -
		f[z][y][x][14] ) / rho;

	FLOAT uz = (
		f[z][y][x][7]  + f[z][y][x][9]  + f[z][y][x][12] +
		f[z][y][x][15] + f[z][y][x][17] - f[z][y][x][3]  -
		f[z][y][x][6]  - f[z][y][x][8]  - f[z][y][x][16] -
		f[z][y][x][18] ) / rho;

	FLOAT sqr = 1.5 * (ux*ux + uy*uy + uz*uz);

	/* update node */
	for(int q = 0; q < 19; q++) {
		FLOAT cu =
			ux*d3q19_v[q][0] + uy*d3q19_v[q][1] + uz*d3q19_v[q][2];
		FLOAT eq = rho * d3q19_w[q] *
			(1. + 3. * cu + 4.5 * cu*cu - sqr);
		f[z][y][x][q] *= (1.0 - omega);
		f[z][y][x][q] += omega * eq;
	}

	/* swap */
	for(int q = 1; q <= 9; q++) {
		FLOAT tmp       = f[z][y][x][q];
		f[z][y][x][q]   = f[z][y][x][q+9];
		f[z][y][x][q+9] = tmp;
	}
}

/* ================================================================== */

void stream(block_t f, int x, int y, int z)
{
	for(int q = 1; q <= 9; q++) {
		int next_row = row;
		int next_col = col;
		int next_x   = x + d3q19_v[q][0];
		int next_y   = y + d3q19_v[q][1];
		int next_z   = z + d3q19_v[q][2];

		/* inner borders extend in xz-plane */
		if(next_x < 0)        { next_col--; next_x += NODES_X; }
		if(next_x >= NODES_X) { next_col++; next_x -= NODES_X; }
		if(next_z < 0)        { next_row--; next_z += NODES_Z; }
		if(next_z >= NODES_Z) { next_row++; next_z -= NODES_Z; }

		/* outer borders are not streamed */
		if(next_col < 0)             { continue; }	/* -x */
		else if(next_col >= CORES_X) { continue; }	/* +x */
		if(next_y < 0)               { continue; }	/* -y */
		else if(next_y >= NODES_Y)   { continue; }	/* +y */
		if(next_row < 0)             { continue; }	/* -z */
		else if(next_row >= CORES_Y) { continue; }	/* +z */

		/* f: local block, g: local/remote block */
		block_t *g = (void*)f;
		if(next_row != row || next_col != col) {
			g = e_get_global_address(next_col, next_row, (void*)f);
		}

		/* stream by swapping */
		FLOAT tmp       = f[z][y][x][q+9];
		f[z][y][x][q+9] = (*g)[next_z][next_y][next_x][q];
		(*g)[next_z][next_y][next_x][q] = tmp;
	}
}

/* ================================================================== */

void bulk(block_t f, FLOAT omega)
{
	/* don't touch border nodes */
	for(int z = 1; z < NODES_Z-1; z++)
		for(int y = 1; y < NODES_Y-1; y++)
			for(int x = 1; x < NODES_X-1; x++)
	{
		/* macroscopic */
		FLOAT rho =
			f[z][y][x][0] + f[z][y][x][1] + f[z][y][x][2] +
			f[z][y][x][3] + f[z][y][x][4] + f[z][y][x][5] +
			f[z][y][x][6] + f[z][y][x][7] + f[z][y][x][8] +
			f[z][y][x][9] + f[z][y][x][10] + f[z][y][x][11] +
			f[z][y][x][12] + f[z][y][x][13] + f[z][y][x][14] +
			f[z][y][x][15] + f[z][y][x][16] + f[z][y][x][17] +
			f[z][y][x][18];

		FLOAT ux = (
			f[z][y][x][10] + f[z][y][x][13] + f[z][y][x][14] +
			f[z][y][x][15] + f[z][y][x][16] - f[z][y][x][1] -
			f[z][y][x][4]  - f[z][y][x][5]  - f[z][y][x][6] -
			f[z][y][x][7] ) / rho;

		FLOAT uy = (
			f[z][y][x][5]  + f[z][y][x][11] + f[z][y][x][13] +
			f[z][y][x][17] + f[z][y][x][18] - f[z][y][x][2]  -
			f[z][y][x][4]  - f[z][y][x][8]  - f[z][y][x][9]  -
			f[z][y][x][14] ) / rho;

		FLOAT uz = (
			f[z][y][x][7]  + f[z][y][x][9]  + f[z][y][x][12] +
			f[z][y][x][15] + f[z][y][x][17] - f[z][y][x][3]  -
			f[z][y][x][6]  - f[z][y][x][8]  - f[z][y][x][16] -
			f[z][y][x][18] ) / rho;

		FLOAT sqr = 1.5 * (ux*ux + uy*uy + uz*uz);

		/* update node */
		for(int q = 0; q < 19; q++) {
			FLOAT cu =
				ux*d3q19_v[q][0] +
				uy*d3q19_v[q][1] +
				uz*d3q19_v[q][2];
			FLOAT eq = rho * d3q19_w[q] *
				(1. + 3. * cu + 4.5 * cu*cu - sqr);
			f[z][y][x][q] *= (1.0 - omega);
			f[z][y][x][q] += omega * eq;
		}

		/* stream */
		for(int q = 0; q < 9; q++) {
			int next_x = x + d3q19_v[q][0];
			int next_y = y + d3q19_v[q][1];
			int next_z = z + d3q19_v[q][2];

			FLOAT tmp       = f[z][y][x][q];
			f[z][y][x][q]   = f[z][y][x][q+9];
			f[z][y][x][q+9] = f[next_z][next_y][next_x][q];
			f[next_z][next_y][next_x][q] = tmp;
		}
	}
}

