/* Lattice Boltzmann Implementation */

#include <stdint.h>
#include <string.h>

#include <e-lib.h>

#include "d3q19.h"
#include "../shared.h"

/* shared memory overlay */
volatile shm_t shm SECTION(".shared_dram");

/* statically allocate dummy memory and local block overlay
   to prevent linker from putting stuff in banks 1..3 */
static uint8_t dummy_bank1[8192] USED SECTION(".data_bank1");
static uint8_t dummy_bank2[8192] USED SECTION(".data_bank2");
static uint8_t dummy_bank3[8192] USED SECTION(".data_bank3");
static block_t *block = (void*)0x2000;

/* barrier structures */
volatile e_barrier_t  barriers[CORES];
         e_barrier_t *tgt_bars[CORES];

/* global index variables */
unsigned int row, col, core;

#define READ_TIMER(X) \
	do { \
		times[X] = E_CTIMER_MAX - e_ctimer_stop(E_CTIMER_0); \
		e_ctimer_set(E_CTIMER_0, E_CTIMER_MAX); \
		e_ctimer_start(E_CTIMER_0, E_CTIMER_CLK); \
	} while(0);

int main()
{
	const FLOAT omega = 1.0;
	unsigned times[TIMERS] = {0};

	/* compile-time bombs, for E16G3 */
	BUILD_BUG(NODES * sizeof(node_t) > 24 * 1024);
	BUILD_BUG(NODES_X < 3 || NODES_Y < 3 || NODES_Z < 3);
	BUILD_BUG(CORES_X < 1 || CORES_Y < 1);
	BUILD_BUG(CORES_X > 4 || CORES_Y > 4);

	/* save mesh coordinates */
	e_coords_from_coreid(e_get_coreid(), &col, &row);
	core = row * CORES_X + col;

	/* initialize barrier */
	e_barrier_init(barriers, tgt_bars);

	/* initialize lattice */
	init(*block);

	/* main loop */
	for(int iter = 0; iter < 3000; iter++) {
#if 0
		/* collide all nodes */
		for(int z = 0; z < NODES_Z; z++)
			for(int y = 0; y < NODES_Y; y++)
				for(int x = 0; x < NODES_X; x++)
					collide(*block, x, y, z, omega);

		/* synchronize */
		e_barrier(barriers, tgt_bars);

		/* stream all nodes */
		for(int z = 0; z < NODES_Z; z++)
			for(int y = 0; y < NODES_Y; y++)
				for(int x = 0; x < NODES_X; x++)
					stream(*block, x, y, z);

#else
		/* collide boundaries: top/bottom (Z) */
		for(int y = 0; y < NODES_Y; y++)
			for(int x = 0; x < NODES_X; x++)
		{
			collide(*block, x, y, 0,         omega);
			collide(*block, x, y, NODES_Z-1, omega);
		}

		/* collide boundaries: front/back (Y) */
		for(int z = 1; z < NODES_Z-1; z++)
			for(int x = 0; x < NODES_X; x++)
		{
			collide(*block, x, 0      ,   z, omega);
			collide(*block, x, NODES_Y-1, z, omega);
		}

		/* collide boundaries: left/right (X) */
		for(int z = 1; z < NODES_Z-1; z++)
			for(int y = 1; y < NODES_Y-1; y++)
		{
			collide(*block, 0,         y, z, omega);
			collide(*block, NODES_X-1, y, z, omega);
		}

		/* synchronize */
		e_barrier(barriers, tgt_bars);

		/* collide and stream the bulk */
		bulk(*block, omega);

		/* stream boundaries: top/bottom (Z) */
		for(int y = 0; y < NODES_Y; y++)
			for(int x = 0; x < NODES_X; x++)
		{
			stream(*block, x, y, 0        );
			stream(*block, x, y, NODES_Z-1);
		}

		/* stream boundaries: front/back (Y) */
		for(int z = 1; z < NODES_Z-1; z++)
			for(int x = 0; x < NODES_X; x++)
		{
			stream(*block, x, 0,         z);
			stream(*block, x, NODES_Y-1, z);
		}

		/* stream boundaries: left/right (X) */
		for(int z = 1; z < NODES_Z-1; z++)
			for(int y = 1; y < NODES_Y-1; y++)
		{
			stream(*block, 0,         y, z);
			stream(*block, NODES_X-1, y, z);
		}
#endif

		/* copy data to shm if necessary */
		if(!(iter % 3)) {
			/* copy lattice to shm */
			memcpy(&shm.lattice[row][col], block, sizeof(block_t));

			/* copy times to shm */
			memcpy(&shm.times[row][col], times, sizeof(times_t));

			/* synchronize */
			e_barrier(barriers, tgt_bars);

			/* core 0: write counter and flag host; wait */
			if(core == 0) {
				shm.iteration = iter;
				shm.pollflag  = POLL_READY;
				while(shm.pollflag == POLL_READY);
			}
		}

		/* synchronize */
		e_barrier(barriers, tgt_bars);
	}

	/* core 0: flag host, all iterations done */
	if(core == 0) {
		shm.pollflag = POLL_DONE;
	}

	/* stop core */
	while(1) {
		__asm__ volatile("idle");
	}
}

