/* Helper Functions to handle data (3D) */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>

#include "../shared.h"

/* fix file owner if run with sudo */
void fixsudo(const char *filename)
{
	if(getenv("SUDO_UID") && getenv("SUDO_GID")) {
		int uid = atoi(getenv("SUDO_UID"));
		int gid = atoi(getenv("SUDO_GID"));
		if(chown(filename, uid, gid)) {
			perror("fixsudo/chown");
			return;
		}
	}
}

#if 0
/* write a (semi-) human-readable dump of the lattice */
void write_populations(d2q9_block_t lattice[CORES_Y][CORES_X], int iter)
{
	FILE *file = fopen("populations.dat", "a");
	if(!file) {
		perror("write_populations/fopen");
		return;
	}

	for(int cy = 0; cy < CORES_Y; cy++) {
		for(int y = 0; y < BLOCKS_Y; y++) {
			for(int cx = 0; cx < CORES_X; cx++) {
				for(int x = 0; x < BLOCKS_X; x++) {
					fprintf(file, "%3d: [%3d,%3d]: ",
						iter,
						cx * BLOCKS_X + x,
						cy * BLOCKS_Y + y
					);
					for(int q = 0; q < 9; q++) {
						fprintf(file, "%.5f ", lattice[cy][cx][y][x][q]);
					}
					fprintf(file, "\n");
				}
			}
		}
	}
	fprintf(file, "\n");

	/* close */
	fclose(file);

	return;
}
#endif

/* write 8-bit grayscale PPM of particle density in a slice */
void write_density(block_t f[CORES_Y][CORES_X], int iter)
{
	char name[32];
	FLOAT min = 1.0, max = 0.0;
	FLOAT rhos[LATTICE_Z][LATTICE_Y][LATTICE_X];

	/* write file header */
	snprintf(name, 32, "./tmp/i%06d.ppm", iter);
	FILE *file = fopen(name, "wb");
	if(!file) {
		perror("write_density/fopen");
		return;
	}
	fprintf(file, "P5\n%d %d\n%d\n", LATTICE_X, LATTICE_Z, 255);

	/* calculate densities from grid */
	for(int cy = 0; cy < CORES_Y; cy++)
		for(int z = 0; z < NODES_Z; z++)
			for(int y = 0; y < NODES_Y; y++)
				for(int cx = 0; cx < CORES_X; cx++)
					for(int x = 0; x < NODES_X; x++)
	{
		const int idx[3] = { cx * NODES_X + x, y, cy * NODES_Z + z};

		FLOAT rho = 0;
		for(int q = 0; q < 19; q++)
			rho += f[cy][cx][z][y][x][q];

		if(rho < min) min = rho;
		if(rho > max) max = rho;
		rhos[idx[2]][idx[1]][idx[0]] = rho;
	}

	/* write slice of grid */
	for(int z = 0; z < LATTICE_Z; z++) {
		for(int x = 0; x < LATTICE_X; x++) {
			FLOAT value = rhos[z][NODES_Y/2][x];

			unsigned char gray = ( 255 *
				(value - min) / (max - min) );
			fwrite(&gray, 1, 1, file);
		}
	}

	/* close the file and fix permissions */
	fclose(file);
	fixsudo(name);

	return;
}

/* write 8-bit grayscale PPM of absolute particle velocity in a slice */
void write_velocity(block_t f[CORES_Y][CORES_X], int iter)
{
	char name[32];
	FLOAT min = 1.0, max = 0.0;
	FLOAT vs[LATTICE_Z][LATTICE_Y][LATTICE_X];

	/* write file header */
	snprintf(name, 32, "./tmp/i%06d.ppm", iter);
	FILE *file = fopen(name, "wb");
	if(!file) {
		perror("write_density/fopen");
		return;
	}
	fprintf(file, "P5\n%d %d\n%d\n", LATTICE_X, LATTICE_Z, 255);

	/* calculate velocities from grid */
	for(int cy = 0; cy < CORES_Y; cy++)
		for(int z = 0; z < NODES_Z; z++)
			for(int y = 0; y < NODES_Y; y++)
				for(int cx = 0; cx < CORES_X; cx++)
					for(int x = 0; x < NODES_X; x++)
	{
		const int idx[3] = { cx * NODES_X + x, y, cy * NODES_Z + z};

		FLOAT rho = 0;
		for(int q = 0; q < 19; q++)
			rho += f[cy][cx][z][y][x][q];

		FLOAT ux = (
			f[cy][cx][z][y][x][10] + f[cy][cx][z][y][x][13] +
			f[cy][cx][z][y][x][14] + f[cy][cx][z][y][x][15] +
			f[cy][cx][z][y][x][16] - f[cy][cx][z][y][x][1]  -
			f[cy][cx][z][y][x][4]  - f[cy][cx][z][y][x][5]  -
			f[cy][cx][z][y][x][6]  - f[cy][cx][z][y][x][7]
			) / rho;

		FLOAT uy = (
			f[cy][cx][z][y][x][5]  + f[cy][cx][z][y][x][11] +
			f[cy][cx][z][y][x][13] + f[cy][cx][z][y][x][17] +
			f[cy][cx][z][y][x][18] - f[cy][cx][z][y][x][2]  -
			f[cy][cx][z][y][x][4]  - f[cy][cx][z][y][x][8]  -
			f[cy][cx][z][y][x][9]  - f[cy][cx][z][y][x][14]
			) / rho;

		FLOAT uz = (
			f[cy][cx][z][y][x][7]  + f[cy][cx][z][y][x][9]  +
			f[cy][cx][z][y][x][12] + f[cy][cx][z][y][x][15] +
			f[cy][cx][z][y][x][17] - f[cy][cx][z][y][x][3]  -
			f[cy][cx][z][y][x][6]  - f[cy][cx][z][y][x][8]  -
			f[cy][cx][z][y][x][16] - f[cy][cx][z][y][x][18]
			) / rho;

		FLOAT u = sqrtf(ux*ux + uy*uy + uz*uz);
		if(u < min) min = u;
		if(u > max) max = u;
		vs[idx[2]][idx[1]][idx[0]] = u;
	}

	/* write slice of grid */
	for(int z = 0; z < LATTICE_Z; z++) {
		for(int x = 0; x < LATTICE_X; x++) {
			FLOAT value = vs[z][NODES_Y/2][x];

			unsigned char gray = ( 255 *
				(value - min) / (max - min) );
			fwrite(&gray, 1, 1, file);
		}
	}

	/* close the file and fix permissions */
	fclose(file);
	fixsudo(name);

	return;
}

/* convert image files to animated gif ./tmp/anim.gif */
void convert_to_gif(void)
{
	int result;

	/* call imagemagick */
	result = system("convert ./tmp/i*.ppm ./tmp/anim.gif"); (void)result;
	fixsudo("./tmp/anim.gif");

	return;
}

/* convert image files to mp4 ./tmp/anim.mp4 */
void convert_to_mp4(void)
{
	int result;

	/* call ffmpeg
	   FIXME: requires ffmpeg binary with '-pattern_type glob' support,
	          so this will not work on the Parallella without updates! */
	result = system("ffmpeg -v quiet -pattern_type glob -i "
	                "'./tmp/i*.ppm' ./tmp/anim.mp4"); (void)result;
	fixsudo("./tmp/anim.mp4");

	return;
}

/* write timer values */
void write_timers(times_t t[CORES_Y][CORES_X], uint32_t iter)
{
	FILE *file = fopen("timers.dat", "ab");
	if(!file) {
		perror("write_timers/fopen");
		return;
	}

	fprintf(file, "Iteration: i=%d\n", iter);
	for(int y = 0; y < CORES_Y; y++) {
		for(int x = 0; x < CORES_X; x++) {
			fprintf(file, "[%d,%d]: ", x, y);
			for(int i = 0; i < TIMERS; i++) {
				fprintf(file, "%8d ", t[y][x][i]);
			}
			fprintf(file, "\n");
		}
	}

	fclose(file);
}

